#!/usr/local/bin/perl -w
#
# sentiment.pl
# 
# CS8761 Final Project (due 12/16/2002 12pm)
#
# Alianza Lima:
#	Amine Abou-Rjeili
#	Suchitra Goopy
#	Amruta Purandare
#	Sam Storie
#
# Description: 
#	This is a program that attempts to assign a sentiment to a review
#	based on the words present. It uses the BigMac and LDOCE machine-
#	readable dictionaries and the "web" to augment them.
#
#	For a complete description of the algorithm used, please see the 
#	accompaning algorithm.txt file.
#
# Files required:
#	DataFiles/allow.pos
#		- Defines the POS tags to allow when filtering. There should be one
#			POS tag per line and it must match the tags used in the BigMac
#			dictionary exactly.
#
#	DataFiles/stop_words
#		- Explicit words to filter out. There should be one word per line
#			and it must match exactly
#
#	DataFiles/LDOCE_Activ_codes.txt
#		- This file contains the sentiment labeled Activ codes from LDOCE. As
#			long as this file is present there should be no modification 
#			required.
#
#	DataFiles/sentimentCache.db
#		- This is a file, in the "DB_File" format, that contains the web
#			cache. If the cache is enabled, then any new words found are
#			stored automatically for future use
#
#
############################################################################
# OPTIONAL EXECUTION PARAMETERS
#
#
# Ability to scale the conribution of words based on their location in the 
#	original review. If you set 1 here, then you must set the SCALE_FUNCTION
#	variable below to either 1, 2 or 3 to have any effect.
#
#	0 = do not scale 
#	1 = scale
#
local $SCALE_VALUES = 0;

# Description of scaling functions
#
#	1 - Words near the end are given more weight in a linear manner
#	2 - Words near beginning are given more weight in a linear manner
#	3 - Words in middle are scaled down but high weight is given to 
#			words near beginning and end
#
#	If you use a value other than these 3 the program will not perform any
#		scaling. Also, for this to take effect you must set SCALE_VALUES to 
#		1 above.
#
local $SCALE_FUNCTION = 1;


# Ability to use a pre-allocated cache to speed up execution. You will probably
#	want to use this unless you want to wait a couple hours per run of the 
#	program. This program caches values in a temporary cache as well, but
#	this temp cache is destroyed when the program exits. Enabling this option
#	allows for a persistant cache that is available across program runs.
#
#	0 = Do not use cache
#	1 = Use cache
#
local $USE_CACHE = 1;


# Use a window when assigning the overall sentiment values to the reviews.
#	If this option is used then the final review sentiments are analyzed
#	and a suitable window is chosen based on the distribution of the sentiment
#	values. The size of this window can be approximately scaled below. However,
#	in this version it will always be centered on 0. There is no option to shift
#	the location of the window at this moment.
#
#	The window size is (by default) calculated by computing the mean of the negative
#	and positive sentiments seperately. Then each value that falls within the range
#	from the negative mean to the positive mean is labeled undecided. This creates
#	a "fuzzy area" that is treated as not conclusive enough to assign a full label.
#
#	1 = Use a window
#	0 = Do not use a window (ie, only sentiments of 0 are 'undecided')
#
local $USE_WINDOW = 1;

# Change the window size by scaling it down by a percentage
#	If this value is set to 1, then about 30% of the reviews will be 
#	classified as 'undecided'. Making it larger than 1 will increase
#	the number of undecided. Setting it between 0 and 1 is probably 
#	most desireable and some example impacts are show below
#
#	1   = About 30% are undecided
#	.5  = About 15% are undecided
#	.25 = About 8% are undecided
#	0   = Only review sentiments of 0 are classified as 'undecided'
#
local $WINDOW_SCALE = 1;


#
# END OF OPTIONAL PARAMETERS
############################################################################




use LDOCE::Interface;
use WebReader::Interface;
use BigMac::QueryData;
use FileHandle;
use DB_File;

my $mainDirectory = shift @ARGV;
die "Usage: sentiment.pl <dir>\n" if !defined $mainDirectory || $mainDirectory eq '';

local $hTrace = new FileHandle;

# Open the trace file
$hTrace->open(">> trace.txt");
$hTrace->autoflush(1);


# Various cache hashes used
#
local %activCodes;
local %wordSenCache;
local %contextSenCache;
local %dbHash;

# Related filenames
#
my $allowPosFn  = 'DataFiles/allow.pos';
my $stopWordsFn = 'DataFiles/stop_words';
my $activCodeFile = 'DataFiles/LDOCE_Activ_codes.txt';
my $dbCacheFile = 'DataFiles/sentimentCache.db';


local( %allowPos )  = ();
local( %stopWords ) = (); # The list of stopwords
local( %synsets )   = (); # The set of synonyms
local( %wordRank )  = (); # The rank of each word according to where in the text it is found

&TRACE("Program execution started.");

&LoadAllowPos( $allowPosFn );
&LoadStopWords( $stopWordsFn );
&InitializeActivCodes($activCodeFile);

if( $USE_CACHE )
{
	&TRACE("Using sentiment value database");
	tie(%dbHash, "DB_File", $dbCacheFile) || die "Error! Can't open sentiment cache.\nReason: $!";
}

if( $USE_WINDOW )
{
	&TRACE("Using a window in final classifications.");
}

&TRACE("Loading file list");
my @fileList = &GetFileList($mainDirectory);
&TRACE("File list loaded.");

&TRACE("Loading modules.");
local $hBigMac = BigMac::QueryData->new;
local $hLdoce = LDOCE::Interface->new;
local $hWebFetch = WebReader::Interface->new;
&TRACE("Modules loaded.");

my( $fmeasure, $precision, $recall );
my( $numCorrect ) = 0;
my( $totalEx ) = 0;
my( $numAttempted ) = 0;
my( $guessSent ) = '';
my @guesses;
my @sents;
my @fileNames;


for $file (@fileList)
{
    # Reset some data structures
    %wordRank = ();
    %synsets  = ();

    # Reset the review sentiment value for this file
    $reviewContext = 0;

    # Split this into the sentiment and the actual filename
    ($sent, $filePath, $fileName) = split /:/, $file;
    
    $totalWords = &FilterText( "$filePath/$fileName" );
    
    foreach $w (keys %wordRank)
	{
		# Now get the context sentiment and we're going to scale it by
		#	the positions of this word
		$sentiment = &DetermineContextSentiment($w);

		# Now tally up the contribution of this word based on it's positions
		#	in the review
		for $pos ( @{ $wordRank{$w} } )
		{
			if( $SCALE_VALUES )
			{
				$reviewContext += &ScaleValue($sentiment, $pos, $totalWords);
			}
			else
			{
				$reviewContext += $sentiment;
			}
		}

	}
	
    push @guesses, $reviewContext;
    push @sents, $sent;
	push @fileNames, $fileName;
	$totalEx++;
}

if( $USE_WINDOW )
{
	@new_guess = &classify(@guesses);
}
else
{
	foreach (0 .. $#guesses)
	{
		if( $guesses[$_] > 0 )
		{
			push @new_guess, "positive";
		}
		elsif( $guesses[$_] < 0 )
		{
			push @new_guess, "negative";
		}
		else
		{
			push @new_guess, "undecided";
		}
	}
}
		

# Now analyze the results
#
foreach (0..$#new_guess)
{
	$numCorrect ++ if $sents[$_] eq $new_guess[$_];
	$numAttempted ++ if $new_guess[$_] ne 'undecided';
	
	&TRACE("$fileNames[$_]\t\t$sents[$_]\t$new_guess[$_]\t$guesses[$_]");
	print "$fileNames[$_]\t\t$sents[$_]\t$new_guess[$_]\n";

}

# All done so calculate the final values
$precision = $numCorrect / $numAttempted;
$recall    = $numCorrect / $totalEx;
$fmeasure = (2 * $recall * $precision) / ($recall + $precision);

printf "%d %d %d\n", $numCorrect, $numAttempted - $numCorrect, $totalEx - $numAttempted;
printf "Precision	=> %0.2f\n", $precision;
printf "Recall		=> %0.2f\n", $recall;
printf "F-measure 	=> %0.2f\n", $fmeasure;


&TRACE("Precision: $precision");
&TRACE("Recall: $recall");
&TRACE("F-measure: $fmeasure");

# Close out the cache file
untie(%dbHash);

# Close the trace file
$hTrace->close;

exit( 0 ); # Done



#
##
#### End of main context
###############################################################
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
###############################################################
#### Subroutines start
##
#


sub GetFileList
{
	my $mainDir = shift;
	my @allFiles = ();
	
	# Now get a listing of all the files in the pos directory
	#
	opendir(POSFILES, "$mainDir/pos") || die "Unable to read $mainDir/pos\nReason: $!\n\n";
	@posFiles = grep /\.txt$/i, readdir POSFILES;	
	
	# Now get a listing of all the files in the neg directory
	#
	opendir(NEGFILES, "$mainDir/neg") || die "Unable to read $mainDir/neg\nReason: $!\n\n";
	@negFiles = grep /\.txt$/i, readdir NEGFILES;	
	
	# Now join these two while tracking the sentiment
	for $file (@posFiles)
	{
		push @allFiles, "positive:$mainDir/pos:$file";
	}

	for $file (@negFiles)
	{
		push @allFiles, "negative:$mainDir/neg:$file";
	}

	return @allFiles;
}


# This will be the subroutine that is used to log actions
sub TRACE
{
	my $string = shift;

	# From Programming Perl:
	my ($sec,$min,$hour) = localtime(time);

	# Make sure the times are formatted pretty
	$sec = $sec < 10 ? '0'.$sec : $sec;
	$min = $min < 10 ? '0'.$min : $min;
	$hour = $hour < 10 ? '0'.$hour : $hour;

	# Print this to the trace file
	print $hTrace "$hour:$min:$sec $string\n";
}



# This function will process the review text
sub FilterText
{
    my( $filename ) = @_;
    my( @headData ) = (); # The head info for a particular word
    my( $accepWord ) = 0; 
    my( $position ) = 0;
    
    open( IN_FILE, $filename ) ||
	die "Error: Could not open file $filename [$!]\n";
    while( <IN_FILE> ){
	chomp;

	# Let us split the line into words with whitespaces as the delimiter        
	$_ = lc; # Convert to lowercase
	s/[^a-z]/ /g; # Filter out everything that is not alpha
	@words = split( /\s+/ );

	foreach $word (@words){
	    $position ++; # Move to the next char
	    next if exists $stopWords{$word}; # Found a stop word

	    $acceptWord = 0; # Do not accept word by default

	    # We will only accept words with a POS in the %allowPos
	    my( @allPos ) = $hBigMac->get_pos( $word );
	    foreach $pos (@allPos){
		if( exists $allowPos{lc $pos} ){
		    $acceptWord = 1;
		    last;
		}
	    }
	    
	    # We could not accept this word, so move to the next one
	    next if $acceptWord == 0;

	    # Now, we will group all the words according to their subject code	    
	    @thesCodes = $hBigMac->get_usage( $word );

	    # Here we need to check if the word has been found or not according to
	    #             @subjectCodes == 0

	    foreach $thesCode (@thesCodes){
		$synsets{$thesCode}{$word} ++;
	    }
	    
	    # Add this position to the existing positions of this word
	    push( @{$wordRank{$word}}, $position );
	}
    }
    close( IN_FILE );

	return $position;
}



###################### Routines to load data from external files
sub LoadStopWords( $ )
{
    my( $filename ) = @_;

    open( INFILE, "$filename" ) || die "Error: Could not open file $filename for stop words\nReason: $!\n";
    while( <INFILE> ){
	chomp;
	
	next if /^\#/; #Ignore comment lines

	$stopWords{$_} = undef;
    }
    close( INFILE );
}


sub LoadAllowPos( $ )
{
    my( $filename ) = @_;

    open( INFILE, "$filename" ) || die "Error: Could not open file $filename for allowed POS\nReason: $!\n";
    while( <INFILE> ){
	chomp;
	
	next if /^\#/; #Ignore comment lines

	$allowPos{$_} = undef;
    }
    close( INFILE );
}




# Initialize the Activ codes present in the data file
#
sub InitializeActivCodes( $ )
{
	my $filename = shift;
	my @label;
	
	open(FILE, "<$filename") || die "Couldn't open file => $filename\nReason: $!\n";

	while( <FILE> )
	{
		chomp;
		@label = split /:/;
		
		# If this code is '+', then it will be assigned a 1 in the first
		#	step, and the second will keep it at one. The same for the case
		#	of '='. If the code is '-' then it will get a 0 in the first step
		#	but a -1 in the second.
		#
		$value = $label[1] eq '+' ? 1 : 0;
		$value = $label[1] eq '-' ? -1 : $value;
	
		$activCodes{$label[0]} = $value;
	}
}





# Determine the sentiment of the context based on a set of words
#	This subroutine also cahces results so that subsequent requests
#	will be much faster
#
sub DetermineContextSentiment
{
	my $string = shift;

	# Perhaps abstract these out to some higher level?
	my $relative = 1;
	my $maxNumber = 10;
	my $numRequests = 10;
	my $minLength = 50;
	my $sleepSeconds = 1;
	my @context = ();
	my $totalSentiment = 0;
	my $wordSentiment = 0;
	my %wordTypes;
	
	# Now check the cache realtive to this execution
	if( defined( $contextSenCache{$string} ) )
	{
		&TRACE("Context sentiment for $string :: $contextSenCache{$string} (cached)");
		return $contextSenCache{$string};
	}
	
	# Check if this word has a sentiment determined from LDOCE
	$wordSentiment = DetermineWordSentiment($string);
	return $wordSentiment if( defined($wordSentiment) );
	
	# Now we need to go to the web and to gather a sentiment value for this
	#	word
	#
	# Check if we're using the file cache
	if( $USE_CACHE && defined( $dbHash{$string} ) )
	{
		&TRACE("Using cached sentences for $string");
		@context = split /|/, $dbHash{$string};
	}
	else
	{
		# Otherwise perform the entire process
		&TRACE("Harvesting sentences for $string");
		@context = $hWebFetch->HarvestSentences($string, $maxNumber, $minLength, $relative, $numRequests, $sleepSeconds);
		
		if( $USE_CACHE )
		{
			# Store this into the sentiment cache
			$dbHash{$string} = join "|", @context;
		}
	}
	

	for $sentence (@context)
	{
		# Split this sentence into it's words and determine the
		#	sentiment of each word present
		$sentence = lc($sentence);
		@words = split / /, $sentence;
		
		for $word (@words)
		{
			# Remove any useless punctuation
			$word =~ s/[^a-z]//g;
			
			# Now hash using the word as the key
			$wordTypes{$word} = 1;
		}
	}

	# Now examine the context for the word types found
	for $word (keys %wordTypes)
	{
		
		$wordSentiment = DetermineWordSentiment($word);
		$totalSentiment += $wordSentiment if( defined($wordSentiment) );
	}


	# Now cache this value for later
	$contextSenCache{$string} = $totalSentiment;
	&TRACE("Context sentiment for $string :: $totalSentiment");

	return $totalSentiment;
}




# Determines the sentiment based on the activ codes present
#	for a word in LDOCE
#
sub DetermineWordSentiment
{
	my $word = shift;

	# If this word has been checked already, then use the cache
	if( defined( $wordSenCache{$word} ) )
	{
		#&TRACE("Word sentiment for $word :: $wordSenCache{$word} (cached)");
		return $wordSenCache{$word};
	}
	
	my $sentiment = 0;
	undef $sentiment;
	
	my @codes = $hLdoce->GetActivCodesForWord( $word );

	if( @codes )
	{
		$sentiment = 0;
		# Process the words and tally the values
		for $code (@codes)
		{
			$sentiment += $activCodes{$code} if( defined($activCodes{$code}) );
		}
	
		#&TRACE("Word sentiment for $word :: $sentiment");
		$wordSenCache{$word} = $sentiment;
	}
	
	# Now scale this down to a 1, 0, or -1 value to prevent words with
	#	several senses from over-powering those with less
	#$sentiment = 1  if( $sentiment > 1 );
	#$sentiment = -1 if( $sentiment < -1 );

	
	return $sentiment;
}



sub ScaleValue( $$$ )
{
	my ($value, $pos, $totalWords) = @_;
	my $scaledValue = 0;
	
	if( $SCALE_FUNCTION == 1 )
	{
		# Scale linearly, based on position, giving more weight to words
		#	near the end of the review
		#
		$scaledValue = ($pos / $totalWords) * $sentiment;
	}
	
	elsif( $SCALE_FUNCTION == 2 )
	{
		# Scale linearly, based on position, giving more weight to words
		#	near the beginning of the review
		#
		$scaledValue = (1 - ($pos / $totalWords)) * $sentiment;
	}

	elsif( $SCALE_FUNCTION == 3 )
	{
		# Scale based on position, but give more weight to words that 
		#	occur near the beginning or end, and give the lest weight
		#	to words in the middle
		#
		$scaledValue = ($pos - ($totalWords / 2)) * 2 / $totalWords;
	}
	else
	{
		$scaledValue = $value;
	}

	&TRACE("Value of $value scaled to $scaledValue");

	return $scaledValue;
}




sub classify
{
	my @doc_sentiments = @_;
	my %freq;                  
	my @final_sent;
	my $total = 0;
	my $lower = 10000000;
	my $upper = -10000000;
	my $negMean = 0;
	my $posMean = 0;
	
	for $sentiment (@doc_sentiments)
	{
		# frequency of the sentiment
		$freq{$sentiment}++;
		$total++;
            
		if($sentiment < $lower)
		{
			$lower = $sentiment;
		}
		elsif($sentiment > $upper)
		{
			$upper = $sentiment;
		}
	}
	
	if( ($lower > 0) && ($upper > 0) )
	{
		# Everything was rated above zero so mark it all positive
		#
		foreach (@doc_sentiments)
		{
			push @final_sent,"positive";
		}
		
		return @final_sent;
	}

	
	if( ($lower < 0) && ($upper < 0) )
	{
		# Everything was rated below zero so mark it all negative
		#
		foreach (@doc_sentiments)
		{ 
			push @final_sent,"negative";
		}
		
		return @final_sent;
	}
		
	for $sentiment (sort {$a <=> $b} keys %freq)
	{
		# If this value was below zero 
		if( $sentiment < 0 )
		{
			# Use it to adjust the "negative" mean
			$negMean += $sentiment * $freq{$sentiment} / $total;
		}
		else
		{
			# Adjust the positive mean
			$posMean += $sentiment * $freq{$sentiment} / $total;
		}
	}
		
	# Comment, SJS
	#
	# The window size could easily be widened or shrunk by using a scaling
	#	factor applied to negMean and posMean here. By multiplying the two means
	#	by a fraction they will converge towards zero, thus shrinking the
	#	window size and increasing the number attempted
	#

	# Added - SJS, 12/16/2002
	#
	# The window size can be scaled by a configurable amount
	$negMean = $negMean * $WINDOW_SCALE;
	$posMean = $posMean * $WINDOW_SCALE;

	&TRACE("Window scaled by a factor of $WINDOW_SCALE");
	&TRACE("Lower bound of undecided window = $negMean");
	&TRACE("Upper bound of undecided window = $posMean");


	foreach (@doc_sentiments)
	{
		if( $_ < $negMean )
		{
			push @final_sent, "negative";
		}
		elsif( $_ > $posMean )
		{
			push @final_sent, "positive";
		}
		else
		{                  
			push @final_sent, "undecided";
		}
	}

	return @final_sent;
}


